//
// ZX Spectrum .SCR screen loader
// Written by Mike Halliday - 2023/2024
//
// For the Einstein Community
//

#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>

#include <graphics.h>

#include <time.h>
#include <conio.h>
#include <sys/ioctl.h> 

// Use MSX video library
#include <msx.h>
#include <msx/gfx.h>

#include <limits.h>

FILE *fpin;

char *buffer, ch;

int file_length;
int block, colourvalues, cellvalue;
int inkcol, papercol, bright;
int screenthird, scanline, characterline, linedata, bufferposition;
int yposition, xposition;
int pixelline, flash, colourvalue;

// Colour table
// ZX Sectrum (x2 including bright)
// 0 - black
// 1 - blue
// 2 - red
// 3 - purple, technically called magenta
// 4 - green
// 5 - pale blue, technically called cyan
// 6 - yellow
// 7 - white
// 8 - transparent (Do not change the paper value in the square being printed)
// 9 - Contrast - currently NOT supported.

// Einstein
// Fixed colour palette (15 in total)
// 0 - transparent
// 1 - black
// 2 - medium green
// 3 - light green
// 4 - dark blue
// 5 - light blue
// 6 - dark red
// 7 - cyan
// 8 - medium red
// 9 - light red
// 10 - dark yellow
// 11 - light yellow
// 12 - dark green
// 13 - magenta
// 14 - grey
// 15 - white

// define Einstein colour translation table
int einsteincol[] = {VDP_INK_BLACK,
                     VDP_INK_DARK_BLUE,
                     VDP_INK_LIGHT_BLUE, 
                     VDP_INK_MAGENTA, 
                     VDP_INK_MEDIUM_GREEN, 
                     VDP_INK_CYAN, 
                     VDP_INK_DARK_YELLOW, 
                     VDP_INK_GRAY, 
                     VDP_INK_TRANSPARENT};
int einsteinbright[] = {VDP_INK_TRANSPARENT,
                        VDP_INK_LIGHT_BLUE,
                        VDP_INK_LIGHT_RED,
                        VDP_INK_MAGENTA,
                        VDP_INK_LIGHT_GREEN,
                        VDP_INK_CYAN,
                        VDP_INK_LIGHT_YELLOW,
                        VDP_INK_WHITE,
                        VDP_INK_TRANSPARENT};

// Custom bit test function
#define btst(num, pos) (num & (1 << pos))

// Custom getbit function
unsigned getbits(unsigned value, unsigned offset, unsigned n)
{
  const unsigned max_n = CHAR_BIT * sizeof(unsigned);
  if (offset >= max_n)
    return 0; /* value is padded with infinite zeros on the left */
  value >>= offset; /* drop offset bits */
  if (n >= max_n)
    return value; /* all  bits requested */
  const unsigned mask = (1u << n) - 1; /* n '1's */
  return value & mask;
}

// Clear console
void myCls()        // Clear the screen (generic or VT100 console)
{
        printf("%c",12);
        printf("\x0c");  // the "\x0c" (0x0c character) resets the cursor position to the top left corner
}

// Start main program
void main(int argc, char *argv[])
{
   if (argc < 2)
   {
      printf("Usage: SCRSHOW <FILENAME.SCR>\n");
      exit(1);
   }

   // Open the file for binary read
   fpin = fopen(argv[1], "rb");
   if (fpin == NULL)
   {
      printf("Failed to open file: %s\n", argv[1]);
      exit(1);
   }

   // Now we try to determine the size of the file to be loaded
   if (fseek(fpin, 0, SEEK_END))
   {
      printf("Couldn't determine size of file\n");
      fclose(fpin);
      exit(1);
   }

   file_length = ftell(fpin);

   if (file_length != 6912)
   {
      printf("Incorrect file size for {%s}\nProbably not a spectrum .scr file\n", argv[1]);
      exit(1);
   }

   // Set file pointer back to start
   fseek(fpin, 0, SEEK_SET);

   // Allocate a buffer to hold the file data
   buffer = (char *)malloc(file_length);
   if (buffer == NULL)
   {
      printf("Failed to allocate memory\n");
      fclose(fpin);
      exit(1);
   }

   // Read the file data into the buffer
   fread(buffer, file_length, 1, fpin);
   fclose(fpin);

   // Read the buffer and draw the pixels using the first 6144 bytes

   // The ZX Spectrum resolution is 256x192 so we need to offset +32x +4y

   // the format is simply 1 bit per pixel (on or off)
   // non linear layout
   // Screen data is split into 3, with 2048 representing 1 3rd of screen
   // Data split across 8 character lines, with all scan lines grouped

   void* bufferstart = (void *) &buffer;

   // Direct screen acess
   vdp_set_mode(mode_1);

   // Clear graphics screen
   clg();

   // 1/3 of a screen at a time
   for (screenthird = 0; screenthird <= 2; screenthird++)
   {
      // Each character line has 8 lines worth of screen data
      for (scanline = 0; scanline <= 7; scanline++)
      {
         // Each character has a part of the screen 
         for (characterline = 0; characterline <= 7; characterline++)
         {
            xposition = 32; // Start offset - removing this causes all sorts of weird screen errors

            // Scan all the 32 lines of characters
            for (linedata = 0; linedata <= 31; linedata++)
            {             
               // Work out the correct set of data from the file to display

               block = (2048 * screenthird); // Add Third offset
               block = block + (scanline * 256);    // Add Character Line
               block = block + (characterline * 32);       // Add scanline
               block = block + linedata;           // Add pixel offset 
               
               yposition = scanline + (characterline * 8) + (screenthird * 64);

               // For each encoded line in the character
               for (pixelline = 7; pixelline >= 0; pixelline--)
               {
                  // Get the colour information starting from the last 768 bytes of the buffer 
                  cellvalue = MODE2_MAX + ((yposition / 8) * 32) + (( xposition / 8 ) - 4);

                  // peek the col value
                  colourvalue = buffer[cellvalue];

                  // calculate the bright value, bit 6 amd set the bright colour if true
                  bright = getbits(colourvalue, 6, 1);
                  if (bright != 0)
                  {
                     // bits 0 to 2
                     inkcol = einsteinbright[getbits(colourvalue, 0, 3)];
                     // bits 3 to 5
                     papercol = einsteinbright[getbits(colourvalue, 3, 3)];
                  }
                  else 
                  {
                     // bits 0 to 2
                     inkcol = einsteincol[getbits(colourvalue, 0, 3)];
                     // bits 3 to 5
                     papercol = einsteincol[getbits(colourvalue, 3, 3)];
                  }
                  
                  // calculate the flash value (Not supported on Einstein)
                  // bit 7
                  flash = getbits(colourvalue, 7, 1);  

                  // Change character colour at current location using direct VDP access
                  vdp_color(inkcol, papercol, 0);  

                  // Plot the pixel if the bit is set, otherwise, background is black
                  if btst(buffer[block],pixelline)
                  {    
                     // plot a pixel in default colour/white
                     plot((xposition + 7) - (pixelline + 32), yposition + 4);
                  }
                  else
                  {
                     // unplot a pixel in default colour/white
                     unplot((xposition + 7) - (pixelline + 32), yposition + 4);
                  }
               } // pixelline
               xposition = xposition + 8;
            } // linedata
         } // characterline
         add_raster_int(1);
      } // scanline
   } // screenthird
   // Wait for keypress
   getchar();
   // clear screen and return to tet mode
   //mode = MODE_TEXT;  // standard text mode
   //console_ioctl(IOCTL_GENCON_SET_MODE, &mode);

   // Direct Screen Access
   vdp_color(VDP_INK_WHITE, VDP_INK_DARK_BLUE, VDP_INK_DARK_BLUE);
   vdp_set_mode(mode_0);
   clg();
   //myCls();        // Clear the screen
}